#  Copyright 2022, The Cozo Project Authors.
#
#  This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
#  If a copy of the MPL was not distributed with this file,
#  You can obtain one at https://mozilla.org/MPL/2.0/.

from pycozo.ext_impl import CozoMagics


def _auto_cozo_mode(lines):
    if lines:
        if lines[0].strip() == '%%py':
            del lines[0]
        elif not (lines[0].startswith('%') or lines[0].startswith('!')):
            lines.insert(0, '%%cozo')
    return lines


def load_ipython_extension(ipython):
    ipython.input_transformers_cleanup.append(_auto_cozo_mode)
    ipython.register_magics(CozoMagics(ipython))



#  Copyright 2022, The Cozo Project Authors.
#
#  This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
#  If a copy of the MPL was not distributed with this file,
#  You can obtain one at https://mozilla.org/MPL/2.0/.
from pycozo import Client
from pycozo.client import QueryException


def test_client():
    client = Client(dataframe=False)

    collected = []

    def cb(a, b, c):
        collected.append((a, b, c))

    inputs = None
    options = None

    def rule_impl(i, o):
        nonlocal inputs, options
        inputs = i
        options = o
        return [('Nicely',), ('Done!',)]

    cb_id = client.register_callback('test', cb)
    client.register_fixed_rule('Holy', 1, rule_impl)
    client.run("?[a, b, c] <- [[1,2,3]] :create test {a => b, c}")
    r = client.run("""
        rel[u, v, w] <- [[1,2,3],[4,5,6]]
        ?[] <~ Holy(rel[], x: 1, y: null)
    """)
    assert r['rows'] == [['Done!'], ['Nicely']]
    client.unregister_callback(cb_id)
    client.unregister_fixed_rule('Holy')
    assert collected == [('Put', [[1, 2, 3]], [])]
    assert inputs == [[[1, 2, 3], [4, 5, 6]]]
    assert options == {'x': 1, 'y': None}
    exported = client.export_relations(['test'])
    assert exported['test']['rows'] == [[1, 2, 3]]
    raised = False
    try:
        client.run("""
                rel[u, v, x] <- [[1,2,3],[4,5,6]]
                
                
                ?[] <~ Holy(rel[], x: 1, y: null)
            """)
    except QueryException:
        raised = True
    assert raised

    data = b'abcxyz'
    r = client.run("?[z] <- [[$z]]", {'z': data})
    assert r['rows'][0][0] == data

    tx = client.multi_transact(True)
    tx.run(':create a {a}')
    tx.run('?[a] <- [[1]] :put a {a}')
    try:
        tx.run(':create a {a}')
    except:
        pass
    tx.run('?[a] <- [[2]] :put a {a}')
    tx.run('?[a] <- [[3]] :put a {a}')
    tx.commit()
    r = client.run('?[a] := *a[a]')
    assert r['rows'] == [[1], [2], [3]]


if __name__ == '__main__':
    test_client()



#  Copyright 2022, The Cozo Project Authors.
#
#  This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
#  If a copy of the MPL was not distributed with this file,
#  You can obtain one at https://mozilla.org/MPL/2.0/.

import json

from IPython.core.magic import (Magics, magics_class, line_magic,
                                cell_magic, needs_local_scope)

from pycozo.client import Client, QueryException


@magics_class
class CozoMagics(Magics):
    """Cozo magics for Jupyter notebooks"""

    def __init__(self, shell, **kwargs):
        super().__init__(shell, **kwargs)
        self.client = None
        self.params = {}

    def _ensure_client(self):
        if self.client is None:
            self.client = Client()
            self.shell.user_ns['COZO_CLIENT'] = self.client

    @line_magic
    def cozo_open(self, line: str):
        """Open a database client. The magic should be followed by two or three arguments:
        ```
        %cozo_open <ENGINE>, <PATH>
        ```
        or
        ```
        %cozo_open <ENGINE>, <PATH>, <OPTIONS>
        ```
        """
        args = eval(line)
        engine = args[0]
        path = args[1]
        options = args[2] if len(args) > 2 else {}
        self.client = Client(engine, path, options)
        self.shell.user_ns['COZO_CLIENT'] = self.client

    @cell_magic
    def cozo(self, line, cell):
        """Run CozoScript"""
        self._ensure_client()
        try:
            res = self.client.run(cell, self.params)
        except QueryException as e:
            return e
        except Exception as e:
            return e
        var = line.strip()
        if var:
            self.shell.user_ns[var] = res
        try:
            return res.style.applymap(_colour_code_type)
        except:
            return res

    @line_magic
    @needs_local_scope
    def cozo_run_string(self, line, local_ns):
        """Run CozoScript contained in a string expression"""
        self._ensure_client()
        script = eval(line, self.shell.user_ns, local_ns)
        if not isinstance(script, str):
            raise Exception('a string is required')
        try:
            return self.client.run(script, self.params)
        except QueryException as e:
            return e

    @line_magic
    def cozo_run_file(self, line):
        """Run CozoScript contained in a file"""
        self._ensure_client()
        filename = eval(line)
        with open(filename, encoding='utf-8') as f:
            script = f.read()
        try:
            return self.client.run(script, self.params)
        except QueryException as e:
            return e

    @line_magic
    def cozo_clear(self, _line):
        """Clear all parameters"""
        self.params.clear()

    @line_magic
    @needs_local_scope
    def cozo_set(self, line, local_ns):
        """Set a parameter
        ```
        %cozo_set <PARAM_NAME> <PARAM_EXPR>
        ```
        """
        var, val = line.split(maxsplit=1)
        val = eval(val, self.shell.user_ns, local_ns)
        try:
            import pandas
            if isinstance(val, pandas.DataFrame):
                self.params[var] = val.values.tolist()
            else:
                self.params[var] = val
        except ImportError:
            self.params[var] = val

    @line_magic
    @needs_local_scope
    def cozo_set_params(self, line, local_ns):
        """Set all parameters, takes a dict"""
        val = eval(line, self.shell.user_ns, local_ns)
        if isinstance(val, dict):
            self.params = val
        else:
            raise Exception('params must be specified as a dict')

    @line_magic
    def cozo_params(self, _line):
        """Returns the current parameters"""
        return self.params

    @line_magic
    def cozo_import_local_file(self, line):
        """Import data saved in a local file"""
        self._ensure_client()
        file = eval(line)
        with open(file, encoding='utf-8') as f:
            data = json.load(f)

        self.client.import_relations(data)


    @line_magic
    def cozo_import_remote_file(self, line):
        """Import data saved in a local file"""
        import requests
        self._ensure_client()
        url = eval(line)
        resp = requests.get(url)
        resp.raise_for_status()
        data = resp.json()
        self.client.import_relations(data)

def _colour_code_type(val):
    if isinstance(val, int) or isinstance(val, float):
        colour = '#307fc1'
    elif isinstance(val, str):
        colour = 'black'
    else:
        colour = '#bf5b3d'
    return f'color: {colour}'



#  Copyright 2023, The Cozo Project Authors.
#
#  This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
#  If a copy of the MPL was not distributed with this file,
#  You can obtain one at https://mozilla.org/MPL/2.0/.

from .builder import *


def test_build():
    rules = [
        ConstantRule(RuleHead('parent', []),
                     Const([['abraham', 'isaac'],
                            ['isaac', 'jakob'],
                            ['jakob', 'joseph']])),
        InlineRule(RuleHead('grandparent', ['A', 'C']),
                   [RuleApply('parent', ['A', 'B']), RuleApply('parent', ['B', 'C'])], ),
        InlineRule(RuleHead('great_grandparent', ['A', 'D']),
                   [RuleApply('parent', ['A', 'B']), RuleApply('parent', ['B', 'C']), RuleApply('parent', ['C', 'D'])]),
        InlineRule(RuleHead('?', ['who']),
                   [RuleApply('great_grandparent', [Const('abraham'), 'who'])])
    ]
    program = InputProgram(rules, limit=10)
    print(program)
    print('---')

    rules = [
        ConstantRule(RuleHead('r1', []), Const([[1, 'a'], [2, 'b']])),
        ConstantRule(RuleHead('r2', []), Const([[2, 'B'], [3, 'C']])),
        InlineRule(RuleHead('?', ['l1', 'l2']), [
            RuleApply('r1', ['a', 'l1']),
            RuleApply('r2', ['a', 'l2'])
        ])
    ]
    program = InputProgram(rules, store_relation=(StoreOp.CREATE, InputRelation('stored', ['l1', 'l2'])))
    print(program)
    print('---')

    rules = [
        InlineRule(RuleHead('?', ['a', 'b']), [
            StoredRuleNamedApply('stored', {'l2': 'b', 'l1': 'a'}),
        ])
    ]
    program = InputProgram(rules)
    print(program)
    print('---')

    rules = [
        ConstantRule(RuleHead('?', []), InputList([
            InputList([
                OpApply('add', [Const(1), Const(2)]),
                OpApply('div', [Const(3), Const(4)]),
                OpApply('eq', [Const(5), Const(6)]),
                OpApply('gt', [Const(7), Const(8)]),
                OpApply('or', [Const(True), Const(False)]),
                OpApply('lowercase', [Const('HELLO')]),
                OpApply('rand_float'),
                OpApply('union', [Const([1, 2, 3]), Const([3, 4, 5]), Const([5, 6, 7])]),
            ]),
        ])),
    ]
    program = InputProgram(rules)
    print(program)
    print('---')



#  Copyright 2022, The Cozo Project Authors.
#
#  This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
#  If a copy of the MPL was not distributed with this file,
#  You can obtain one at https://mozilla.org/MPL/2.0/.

import json
import logging

logger = logging.getLogger(__name__)


class Client:
    """Python client for CozoDB

    This client can either operate on an embedded database, or a remote database via HTTP.
    """

    def __init__(self, engine='mem', path='', options=None, *, dataframe=True):
        """Constructor for the client. The behaviour depends on the argument.

        If the database `db` is an embedded one, and you do not intend it to live as long as your program, you **must**
        call `db.close()` when you are done with it. Simply `del db` is not enough to clean up the native parts.

        :param engine: if 'http', then a remote client is constructed, otherwise an embedded one is constructed.
                       For 'http', the `requests` package must be installed. For the embedded engines,
                       the `cozo-embedded` package must be installed.
                       What engines can be used depends on what was compiled in. Use 'mem' for in-memory, non-persistent
                       databases, 'sqlite' for lightweight persistent databases, and 'rocksdb' for databases
                       that expects high concurrency.
        :param path: the path to store the database on disk, only makes sense for those engines that are persistent.
        :param options: options for the database, the expected values depend on the engine of the database.
                        Currently only the 'http' engine expect options of the form:
                        `{'host': <HOST:PORT>, 'auth': <AUTH_STR>}`.
        :param dataframe: if true, output will be transformed into pandas dataframes. The `pandas` package
                          must be installed.
        """
        self.pandas = None
        if engine == 'http':
            self.host = options['host']
            self.auth = options.get('auth')
            self.embedded = None
            self._remote_sse = {}
            self._remote_cb_id = 0
        else:
            from cozo_embedded import CozoDbPy
            self.embedded = CozoDbPy(engine, path, json.dumps(options or {}))

        if dataframe:
            try:
                import pandas
                self.pandas = pandas
            except ImportError:
                logger.exception('`pandas` feature was requested, but pandas is not installed')
                pass

    def close(self):
        """Close the embedded database. After closing, the database can no longer be used.

        For embedded databases, this method must be called, otherwise the native resources associated with it
        may live as long as your program.

        This is a no-op for HTTP-based clients.
        """
        if self.embedded:
            self.embedded.close()

    def _headers(self):
        return {
            'x-cozo-auth': self.auth
        }

    def _client_request(self, script, params=None, immutable=False):
        import requests

        r = requests.post(f'{self.host}/text-query', headers=self._headers(), json={
            'script': script,
            'params': params or {},
            'immutable': immutable
        })
        res = r.json()
        return self._format_return(res)

    def _format_return(self, res):
        if not res['ok']:
            raise QueryException(res)

        if self.pandas:
            return self.pandas.DataFrame(columns=res['headers'], data=res['rows'])
        else:
            return res

    def _embedded_request(self, script, params=None, immutable=False):
        try:
            res = self.embedded.run_script(script, params or {}, immutable)
        except Exception as e:
            raise QueryException(e.args[0]) from None
        if self.pandas:
            return self.pandas.DataFrame(columns=res['headers'], data=res['rows'])
        else:
            return res

    def run(self, script, params=None, immutable=False):
        """Run a given CozoScript query.

        :param script: the query in CozoScript
        :param params: the named parameters for the query. If specified, must be a dict with string keys.
        :return: the query result as a dict, or a pandas dataframe if the `dataframe` option was true.
        """
        if self.embedded is None:
            return self._client_request(script, params, immutable)
        else:
            return self._embedded_request(script, params, immutable)

    def export_relations(self, relations):
        """Export the specified relations.

        :param relations: names of the relations in a list.
        :return: a dict with string keys for the names of relations, and values containing all the rows.
        """
        if self.embedded:
            return self.embedded.export_relations(relations)
        else:
            import requests
            import urllib.parse

            rels = ','.join(map(lambda s: urllib.parse.quote_plus(s), relations))
            url = f'{self.host}/export/{rels}'

            r = requests.get(url, headers=self._headers())
            res = r.json()
            if res['ok']:
                return res['data']
            else:
                raise RuntimeError(res['message'])

    def import_relations(self, data):
        """Import data into a database

        Note that triggers are _not_ run for the relations, if any exists.
        If you need to activate triggers, use queries with parameters.

        :param data: should be given as a dict with string keys, in the same format as returned by `export_relations`.
                     The relations to import into must exist.
        """
        if self.embedded:
            self.embedded.import_relations(data)
        else:
            import requests
            url = f'{self.host}/import'

            r = requests.put(url, headers=self._headers(), json=data)
            res = r.json()
            if not res['ok']:
                raise RuntimeError(res['message'])

    def backup(self, path):
        """Backup a database to the specified path.

        :param path: the path to write the backup into. For a remote database, this is a path on the remote machine.
        """
        if self.embedded:
            self.embedded.backup(path)
        else:
            import requests

            r = requests.post(f'{self.host}/backup', headers=self._headers(), json={'path': path})
            res = r.json()
            if not res['ok']:
                raise RuntimeError(res['message'])

    def register_callback(self, relation, callback):
        if self.embedded:
            return self.embedded.register_callback(relation, callback)
        else:
            import threading

            tid = self._remote_cb_id
            self._remote_cb_id += 1
            url = f'{self.host}/changes/{relation}'
            thread = threading.Thread(target=self._start_sse, args=(tid, url, callback), daemon=True)
            thread.start()
            self._remote_sse[tid] = {'thread': thread}

            return tid

    def _start_sse(self, tid, url, callback, min_delay=1, max_delay=60):
        import requests
        logger.info('Starting SSE thread')
        headers = {'Accept': 'text/event-stream', 'Accept-Encoding': ''}

        consecutive_failures = 0

        while True:
            try:
                with requests.get(url, stream=True, headers=headers) as response:
                    response.raise_for_status()

                    consecutive_failures = 0

                    buffer = b""
                    for chunk in response.iter_content(chunk_size=1):
                        if tid not in self._remote_sse:
                            logger.info('Stopping SSE thread')
                            return
                        buffer += chunk
                        if buffer.endswith(b'\n\n'):
                            event_text = buffer.decode('utf-8').strip()
                            if event_text.startswith('data:'):
                                payload = json.loads(event_text[5:].strip())
                                callback(payload['op'], payload['new_rows']['rows'], payload['old_rows']['rows'])
                            buffer = b""
            except Exception as e:
                import time

                logger.error(f'Error in SSE thread: {e}')
                consecutive_failures += 1

                # Exponential backoff with a cap at max_delay
                backoff_delay = min(min_delay * (2 ** (consecutive_failures - 1)), max_delay)

                logger.info(f'Sleeping for {backoff_delay} seconds before retrying...')
                time.sleep(backoff_delay)

    def unregister_callback(self, cb_id):
        if self.embedded:
            self.embedded.unregister_callback(cb_id)
        else:
            del self._remote_sse[cb_id]

    def register_fixed_rule(self, name, arity, impl):
        if self.embedded:
            return self.embedded.register_fixed_rule(name, arity, impl)
        else:
            raise RuntimeError('Only supported on embedded DBs')

    def unregister_fixed_rule(self, name):
        if self.embedded:
            return self.embedded.unregister_fixed_rule(name)
        else:
            raise RuntimeError('Only supported on embedded DBs')

    def restore(self, path):
        """Restore database from a backup. Must be called on an empty database.

        :param path: the path to the backup.
                     For remote databases, you cannot restore them this way: use the executable directly.
        """
        if self.embedded:
            self.embedded.restore(path)
        else:
            raise RuntimeError('Remote databases cannot be restored remotely')

    def import_from_backup(self, path, relations):
        """Import stored relations from a backup.

        Note that triggers are _not_ run for the relations, if any exists.
        If you need to activate triggers, use queries with parameters.

        :param path: path to the backup file. For remote databases, this is a path on the remote machine.
        :param relations: a list containing the names of the relations to import. The relations must exist
                          in the database.
        """
        if self.embedded:
            self.embedded.import_from_backup(path, relations)
        else:
            import requests

            r = requests.post(f'{self.host}/import-from-backup', headers=self._headers(),
                              json={'path': path, 'relations': relations})
            res = r.json()
            if not res['ok']:
                raise RuntimeError(res['message'])

    def multi_transact(self, write=False):
        if self.embedded:
            return MultiTransact(self.embedded.multi_transact(write))
        else:
            raise RuntimeError('Multi-transaction not yet supported for remote')

    def _process_mutate_data_dict(self, data):
        cols = []
        row = []
        for k, v in data.items():
            cols.append(k)
            row.append(v)
        return cols, row

    def _process_mutate_data(self, data):
        if isinstance(data, dict):
            cols, row = self._process_mutate_data_dict(data)
            return ','.join(cols), [row]
        elif isinstance(data, list):
            cols, row = self._process_mutate_data_dict(data[0])
            rows = [row]
            for el in data[1:]:
                nxt_row = []
                for col in cols:
                    nxt_row.append(el[col])
                rows.append(nxt_row)
            return ','.join(cols), rows
        else:
            import pandas as pd
            if isinstance(data, pd.DataFrame):
                cols = data.columns.tolist()
                rows = data.values.tolist()
                return ','.join(cols), rows
            else:
                raise RuntimeError('Invalid data type for mutation')

    def _mutate(self, relation, data, op):
        cols_str, processed_data = self._process_mutate_data(data)
        q = f'?[{cols_str}] <- $data :{op} {relation} {{ {cols_str} }}'
        return self.run(q, {'data': processed_data})

    def insert(self, relation, data):
        return self._mutate(relation, data, 'insert')

    def put(self, relation, data):
        return self._mutate(relation, data, 'put')

    def update(self, relation, data):
        return self._mutate(relation, data, 'update')

    def rm(self, relation, data):
        return self._mutate(relation, data, 'rm')


class MultiTransact:
    def __init__(self, multi_tx):
        self.multi_tx = multi_tx

    def commit(self):
        return self.multi_tx.commit()

    def abort(self):
        return self.multi_tx.abort()

    def run(self, script, params=None):
        return self.multi_tx.run_script(script, params or {})


class QueryException(Exception):
    """The exception class for queries. `repr(e)` will pretty format the exceptions into ANSI-coloured messages.
    """

    def __init__(self, resp):
        super().__init__()
        self.resp = resp

    def __repr__(self):
        if hasattr(self.resp, 'get'):
            return self.resp.get('display') or self.resp.get('message') or str(self.resp)
        else:
            return str(self.resp)

    def __str__(self):
        return self.resp.get('message') or str(self.resp)

    def _repr_pretty_(self, p, cycle):
        p.text(repr(self))

    @property
    def code(self):
        return self.resp.get('code')



#  Copyright 2022, The Cozo Project Authors.
#
#  This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
#  If a copy of the MPL was not distributed with this file,
#  You can obtain one at https://mozilla.org/MPL/2.0/.

from pycozo.client import Client



#  Copyright 2022, The Cozo Project Authors.
#
#  This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
#  If a copy of the MPL was not distributed with this file,
#  You can obtain one at https://mozilla.org/MPL/2.0/.

from setuptools import setup

VERSION = '0.7.2'

setup(
    name='pycozo',
    version=VERSION,
    packages=['pycozo'],
    url='',
    license='MPL-2.0',
    author='Ziyang Hu',
    author_email='hu.ziyang@cantab.net',
    description='Python client for the Cozo database',
    install_requires=[],
    extras_require={
        'pandas': ['pandas', 'ipython'],
        'embedded': ['cozo-embedded==' + VERSION],
        'client': ['requests']
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Framework :: IPython",
        "Programming Language :: Python",
        "Topic :: Utilities",
    ],
)



#  Copyright 2023, The Cozo Project Authors.
#
#  This Source Code Form is subject to the terms of the Mozilla Public License, v. 2.0.
#  If a copy of the MPL was not distributed with this file,
#  You can obtain one at https://mozilla.org/MPL/2.0/.
import json
from enum import Enum
from dataclasses import dataclass, field
from typing import Any, TypeAlias, Union


@dataclass
class Var:
    name: str

    def __str__(self):
        return self.name


@dataclass
class Const:
    value: Any

    def __str__(self):
        return json.dumps(self.value)


@dataclass
class InputParam:
    name: str

    def __str__(self):
        return '$' + self.name


Expr: TypeAlias = Union[str | Var, Const, InputParam, 'InputObject', 'InputList', 'OpApply']


@dataclass
class InputObject:
    kvs: dict[str, Expr]

    def __str__(self):
        return '{' + ', '.join(f'{k}: {v}' for k, v in self.kvs) + '}'


@dataclass
class InputList:
    items: list[Expr]

    def __str__(self):
        return '[' + ', '.join(map(str, self.items)) + ']'


@dataclass
class OpApply:
    op: str
    args: list[Expr] = field(default_factory=list)

    def __str__(self):
        return f'{self.op}({", ".join(map(str, self.args))})'


class StoreOp(Enum):
    CREATE = 'create'
    REPLACE = 'replace'
    PUT = 'put'
    UPDATE = 'update'
    RM = 'rm'
    ENSURE = 'ensure'
    ENSURE_NOT = 'ensure_not'

    def __str__(self):
        return ':' + self.value


@dataclass
class Sorter:
    column: str
    aggr: str | None = None
    reverse: bool = False

    def __str__(self):
        ret = ""
        if self.reverse:
            ret += "-"
        if self.aggr:
            ret += f"{self.aggr}("
        ret += self.column
        if self.aggr:
            ret += ")"
        return ret


@dataclass
class RuleHead:
    name: str
    args: list[str | tuple[str, str]]

    def __str__(self):
        ret = self.name + "["
        for arg in self.args:
            if isinstance(arg, tuple):
                ret += f"{arg[0]}({arg[1]}), "
            else:
                ret += f"{arg}, "
        if len(self.args) > 0:
            ret = ret[:-2]
        ret += "]"
        return ret


@dataclass
class InputRelation:
    name: str
    keys: list[str]
    values: list[str] = field(default_factory=list)

    def __str__(self):
        ret = self.name + ' {'
        for k in self.keys:
            ret += f'{k}, '

        if len(self.values) > 0:
            ret += ' => '
            for v in self.values:
                ret += f'{v}, '
        if ret.endswith(', '):
            ret = ret[:-2]
        ret += ' }'
        return ret


@dataclass
class RuleApply:
    name: str
    args: list[Expr] = field(default_factory=list)

    def __str__(self):
        return f'{self.name}[{", ".join(map(str, self.args))}]'


@dataclass
class StoredRuleApply:
    name: str
    args: list[Expr] = field(default_factory=list)
    validity: Expr | None = None

    def __str__(self):
        return f'*{self.name}[{", ".join(map(str, self.args))} {"| @" + str(self.validity) if self.validity else ""}]'


@dataclass
class StoredRuleNamedApply:
    name: str
    args: dict[str, Expr]
    validity: Expr | None = None

    def __str__(self):
        ret = '*' + self.name + '{'
        for k, v in self.args.items():
            ret += f'{k}: {v}, '
        if self.validity:
            ret += f'| @{self.validity}'
        if ret.endswith(', '):
            ret = ret[:-2]
        ret += '}'
        return ret


@dataclass
class ProximityApply:
    name: str
    args: dict[str, Expr]
    params: dict[str, Expr]

    def __str__(self):
        ret = '~' + self.name + '{'
        for k, v in self.args.items():
            ret += f'{k}: {v}, '
        ret += '| '
        for k, v in self.params.items():
            ret += f'{k}: {v}, '
        ret += '}'
        return ret


@dataclass
class Conjunction:
    atoms: list['Atom']

    def __str__(self):
        ret = ''
        for part in self.atoms:
            if ret != '':
                ret += ', (' + str(part) + ')'
            else:
                ret += '(' + str(part) + ')'


@dataclass
class Disjunction:
    atoms: list['Atom']

    def __str__(self):
        ret = ''
        for part in self.atoms:
            if ret != '':
                ret += ' or (' + str(part) + ')'
            else:
                ret += '(' + str(part) + ')'


@dataclass
class Negation:
    atom: 'Atom'

    def __str__(self):
        return 'not (' + str(self.atom) + ')'


@dataclass
class Bind:
    name: str
    expr: Expr
    multi_bind: bool = False

    def __str__(self):
        return f'{self.name} {"in" if self.multi_bind else "="} {self.expr}'


@dataclass
class Cond:
    clauses: list[(Expr, Expr)]

    def __str__(self):
        return 'cond ' + ', '.join(f'{k}, {v}, ' for k, v in self.clauses)


@dataclass
class RawAtom:
    script: str

    def __str__(self):
        return f'({self.script})'


Atom: TypeAlias = Expr | Cond | Bind | RuleApply | StoredRuleApply | StoredRuleNamedApply | Conjunction | Disjunction | Negation | RawAtom


@dataclass
class FixedRule:
    head: RuleHead
    rule_name: str
    inputs: list[RuleApply | StoredRuleApply | StoredRuleNamedApply] = field(default_factory=list)
    parameters: dict[str, Expr] = field(default_factory=dict)

    def __str__(self):
        ret = f'{self.head} <~\n    {self.rule_name}('
        for rule in self.inputs:
            ret += str(rule) + ', '
        for k, v in self.parameters.items():
            ret += f'{k}: {v}, '
        ret += ')\n'
        return ret


@dataclass
class InlineRule:
    head: RuleHead
    atoms: list[Atom]

    def __str__(self):
        ret = f'{self.head} :=\n'
        for atom in self.atoms[:-1]:
            ret += '    ' + str(atom) + ',\n'
        ret += '    ' + str(self.atoms[-1]) + '\n'
        return ret


@dataclass
class ConstantRule:
    head: RuleHead
    body: Expr

    def __str__(self):
        return f'{self.head} <-\n    {self.body}\n'


@dataclass
class InputProgram:
    rules: list[ConstantRule | InlineRule | FixedRule]
    limit: int | None = None
    offset: int | None = None
    sorters: list[Sorter] = field(default_factory=list)
    store_relation: tuple[StoreOp, InputRelation] | None = None

    def __str__(self):
        ret = ''
        for rule in self.rules:
            ret += str(rule)
        ret += '\n'
        if self.limit:
            ret += f':limit {self.limit}\n'
        if self.offset:
            ret += f':offset {self.offset}\n'
        if len(self.sorters) > 0:
            ret += ':sort ' + ', '.join(map(str, self.sorters)) + '\n'
        if self.store_relation:
            ret += f'{self.store_relation[0]} {self.store_relation[1]}' + '\n'
        return ret.strip()
